"""Functional implementation of spiking layers."""

import torch


class SpikingActivation(torch.autograd.Function):
    """
    Function for converting an arbitrary activation function to a spiking equivalent.

    Notes
    -----
    We would not recommend calling this directly, use
    `pytorch_spiking.SpikingActivation` instead.
    """

    @staticmethod
    def forward(
        ctx,
        inputs,
        activation,
        dt=0.001,
        initial_state=None,
        spiking_aware_training=True,
        return_sequences=False,
        training=False,
    ):
        """
        Forward pass of SpikingActivation function.

        Parameters
        ----------
        inputs : ``torch.Tensor``
            Array of input values with shape ``(batch_size, n_steps, n_neurons)``.
        activation : callable
            Activation function to be converted to spiking equivalent.
        dt : float
            Length of time (in seconds) represented by one time step.
        initial_state : ``torch.Tensor``
            Initial spiking voltage state (should be an array with shape
            ``(batch_size, n_neurons)``, with values between 0 and 1). Will use a
            uniform distribution if none is specified.
        spiking_aware_training : bool
            If True (default), use the spiking activation function
            for the forward pass and the base activation function for the backward pass.
            If False, use the base activation function for the forward and
            backward pass during training.
        return_sequences : bool
            Whether to return the last output in the output sequence (default), or the
            full sequence.
        training : bool
            Whether this function should be executed in training or evaluation mode
            (this only matters if ``spiking_aware_training=False``).
        """

        ctx.activation = activation
        ctx.return_sequences = return_sequences
        ctx.save_for_backward(inputs)

        if training and not spiking_aware_training:
            output = activation(inputs if return_sequences else inputs[:, -1])
            return output

        if initial_state is None:
            initial_state = torch.rand(
                inputs.shape[0], inputs.shape[2], dtype=inputs.dtype
            )

        # match inputs to initial state dtype if one was passed in
        inputs = inputs.type(initial_state.dtype)

        voltage = initial_state
        all_spikes = []
        rates = activation(inputs) * dt
        for i in range(inputs.shape[1]):
            voltage += rates[:, i]
            n_spikes = torch.floor(voltage)
            voltage -= n_spikes
            if return_sequences:
                all_spikes.append(n_spikes)

        if return_sequences:
            output = torch.stack(all_spikes, dim=1)
        else:
            output = n_spikes

        output /= dt

        return output

    @staticmethod
    def backward(ctx, grad_output):
        """Backward pass of SpikingActivation function."""

        # TODO: is there a way to reuse the forward pass activations computed in
        #  `forward`? the below results in an infinite loop
        # inputs, rates = ctx.saved_tensors
        # return torch.autograd.grad(rates, inputs, grad_outputs=grad_output)

        inputs = ctx.saved_tensors[0]
        with torch.enable_grad():
            output = ctx.activation(inputs if ctx.return_sequences else inputs[:, -1])
            return (
                torch.autograd.grad(output, inputs, grad_outputs=grad_output)
                + (None,) * 7
            )


spiking_activation = SpikingActivation.apply



"""Modules for adding spiking behaviour to PyTorch models."""

import numpy as np
import torch

from pytorch_spiking.functional import spiking_activation


class SpikingActivation(torch.nn.Module):  # pylint: disable=abstract-method
    """Module for converting an arbitrary activation function to a spiking equivalent.

    Neurons will spike at a rate proportional to the output of the base activation
    function. For example, if the activation function is outputting a value of 10, then
    the wrapped SpikingActivationCell will output spikes at a rate of 10Hz (i.e., 10
    spikes per 1 simulated second, where 1 simulated second is equivalent to ``1/dt``
    time steps). Each spike will have height ``1/dt`` (so that the integral of the
    spiking output will be the same as the integral of the base activation output).
    Note that if the base activation is outputting a negative value then the spikes
    will have height ``-1/dt``. Multiple spikes per timestep are also possible, in
    which case the output will be ``n/dt`` (where ``n`` is the number of spikes).

    When applying this layer to an input, make sure that the input has a time axis.
    The spiking output will be computed along the time axis.
    The number of simulation timesteps will depend on the length of that time axis.
    The number of timesteps does not need to be the same during
    training/evaluation/inference. In particular, it may be more efficient
    to use one timestep during training and multiple timesteps during inference
    (often with ``spiking_aware_training=False``, and ``apply_during_training=False``
    on any `.Lowpass` layers).

    Parameters
    ----------
    activation : callable
        Activation function to be converted to spiking equivalent.
    dt : float
        Length of time (in seconds) represented by one time step.
    initial_state : ``torch.Tensor``
        Initial spiking voltage state (should be an array with shape
        ``(batch_size, n_neurons)``, with values between 0 and 1). Will use a uniform
        distribution if none is specified.
    spiking_aware_training : bool
        If True (default), use the spiking activation function
        for the forward pass and the base activation function for the backward pass.
        If False, use the base activation function for the forward and
        backward pass during training.
    return_sequences : bool
        Whether to return the full sequence of output spikes (default),
        or just the spikes on the last timestep.
    """

    def __init__(
        self,
        activation,
        dt=0.001,
        initial_state=None,
        spiking_aware_training=True,
        return_sequences=True,
    ):
        """"""  # empty docstring removes useless parent docstring from docs
        super().__init__()

        self.activation = activation
        self.initial_state = initial_state
        self.dt = dt
        self.spiking_aware_training = spiking_aware_training
        self.return_sequences = return_sequences

    def forward(self, inputs):
        """Compute output spikes given inputs.

        Parameters
        ----------
        inputs : ``torch.Tensor``
            Array of input values with shape ``(batch_size, n_steps, n_neurons)``.

        Returns
        -------
        outputs : ``torch.Tensor``
            Array of output spikes with shape ``(batch_size, n_neurons)`` if
            ``return_sequences=False`` else ``(batch_size, n_steps, n_neurons)``. Each
            element will have value ``n/dt``, where ``n`` is the number of spikes
            emitted by that neuron on that time step.
        """
        return spiking_activation(
            inputs,
            self.activation,
            self.dt,
            self.initial_state,
            self.spiking_aware_training,
            self.return_sequences,
            self.training,
        )


class Lowpass(torch.nn.Module):  # pylint: disable=abstract-method
    """Module implementing a Lowpass filter.

    The initial filter state and filter time constants are both trainable
    parameters. However, if ``apply_during_training=False`` then the parameters are
    not part of the training loop, and so will never be updated.

    When applying this layer to an input, make sure that the input has a time axis.

    Parameters
    ----------
    tau : float
        Time constant of filter (in seconds).
    dt : float
        Length of time (in seconds) represented by one time step.
    apply_during_training : bool
        If False, this layer will effectively be ignored during training (this
        often makes sense in concert with the swappable training behaviour in, e.g.,
        `.SpikingActivation`, since if the activations are not spiking during training
        then we often don't need to filter them either).
    level_initializer : ``torch.Tensor``
        Initializer for filter state.
    return_sequences : bool
        Whether to return the full sequence of filtered output (default),
        or just the output on the last timestep.
    """

    def __init__(
        self,
        tau,
        units,
        dt=0.001,
        apply_during_training=True,
        initial_level=None,
        return_sequences=True,
    ):
        """"""  # empty docstring removes useless parent docstring from docs
        super().__init__()

        if tau <= 0:
            raise ValueError("tau must be a positive number")

        self.tau = tau
        self.units = units
        self.dt = dt
        self.apply_during_training = apply_during_training
        self.initial_level = initial_level
        self.return_sequences = return_sequences

        # apply ZOH discretization
        smoothing_init = np.exp(-self.dt / self.tau)

        # compute inverse sigmoid of tau, so that when we apply the sigmoid
        # later we'll get the tau value specified
        self.smoothing_init = np.log(smoothing_init / (1 - smoothing_init))

        self.level_var = torch.nn.Parameter(
            torch.zeros(1, units) if self.initial_level is None else self.initial_level
        )

        self.smoothing_var = torch.nn.Parameter(
            torch.ones(1, units) * self.smoothing_init
        )

    def forward(self, inputs):
        """Apply filter to inputs.

        Parameters
        ----------
        inputs : ``torch.Tensor``
            Array of input values with shape ``(batch_size, n_steps, units)``.

        Returns
        -------
        outputs : ``torch.Tensor``
            Array of output spikes with shape ``(batch_size, units)`` if
            ``return_sequences=False`` else ``(batch_size, n_steps, units)``.
        """

        if self.training and not self.apply_during_training:
            return inputs if self.return_sequences else inputs[:, -1]

        level = self.level_var
        smoothing = torch.sigmoid(self.smoothing_var)

        # cast inputs to module type
        inputs = inputs.type(self.smoothing_var.dtype)

        all_levels = []
        for i in range(inputs.shape[1]):
            level = (1 - smoothing) * inputs[:, i] + smoothing * level
            if self.return_sequences:
                all_levels.append(level)

        if self.return_sequences:
            return torch.stack(all_levels, dim=1)
        else:
            return level


class TemporalAvgPool(torch.nn.Module):
    """Module for taking the average across one dimension of a tensor.

    Parameters
    ----------
    dim : int, optional
        The dimension to average across. Defaults to the second dimension (``dim=1``),
        which is typically the time dimension (for tensors that have a time dimension).
    """

    def __init__(self, dim=1):
        """"""  # empty docstring removes useless parent docstring from docs
        super().__init__()
        self.dim = dim

    def forward(self, inputs):
        """Apply average pooling to inputs.

        Parameters
        ----------
        inputs : ``torch.Tensor``
            Array of input values with shape ``(batch_size, n_steps, ...)``.

        Returns
        -------
        outputs : ``torch.Tensor``
            Array of output values with shape ``(batch_size, ...)``.
            The time dimension is fully averaged and removed.
        """
        return torch.mean(inputs, dim=self.dim)



import copy

import numpy as np
import pytest
import torch


from pytorch_spiking import modules


@pytest.mark.parametrize(
    "activation", (torch.nn.ReLU(), torch.nn.functional.relu, torch.tanh)
)
def test_activations(activation, rng, allclose):
    x = torch.from_numpy(rng.randn(32, 10, 2))

    ground = activation(x)

    # behaviour equivalent to base activation during training
    y = modules.SpikingActivation(
        activation, return_sequences=True, spiking_aware_training=False
    ).train()(x)
    assert allclose(y, ground)

    # not equivalent during inference
    y = modules.SpikingActivation(
        activation, return_sequences=True, spiking_aware_training=False
    ).eval()(x)
    assert not allclose(y, ground, record_rmse=False, print_fail=0)

    # equivalent during inference, with large enough dt
    y = modules.SpikingActivation(
        activation, return_sequences=True, spiking_aware_training=False, dt=1e8
    ).eval()(x)
    assert allclose(y, ground)

    # not equivalent during training if using spiking_aware_training
    y = modules.SpikingActivation(
        activation, return_sequences=True, spiking_aware_training=True
    ).train()(x)
    assert not allclose(y, ground, record_rmse=False, print_fail=0)

    # equivalent with large enough dt
    y = modules.SpikingActivation(
        activation,
        return_sequences=True,
        spiking_aware_training=True,
        dt=1e8,
    ).train()(x)
    assert allclose(y, ground)


def test_initial_state(seed, allclose):
    x = torch.from_numpy(np.ones((2, 100, 10)) * 100)

    init = torch.rand((2, 10), generator=torch.random.manual_seed(seed))

    # layers with the same initial state produce the same output
    y0 = modules.SpikingActivation(
        torch.nn.ReLU(), return_sequences=True, initial_state=init
    )(x)
    y1 = modules.SpikingActivation(
        torch.nn.ReLU(), return_sequences=True, initial_state=init
    )(x)
    assert allclose(y0, y1)

    # layers with different initial state produce different output
    y2 = modules.SpikingActivation(torch.nn.ReLU(), return_sequences=True)(x)
    assert not allclose(y0, y2, record_rmse=False, print_fail=0)

    # the same layer called multiple times will produce the same output (if the initial
    # state is set)
    layer = modules.SpikingActivation(
        torch.nn.ReLU(), return_sequences=True, initial_state=init
    )
    assert allclose(layer(x), layer(x))

    # layer will produce different output each time if initial state not set
    layer = modules.SpikingActivation(torch.nn.ReLU(), return_sequences=True)
    assert not allclose(layer(x), layer(x), record_rmse=False, print_fail=0)


def test_spiking_aware_training(rng, allclose):
    layer = modules.SpikingActivation(
        torch.nn.ReLU(), spiking_aware_training=False
    ).train()
    layer_sat = modules.SpikingActivation(
        torch.nn.ReLU(), spiking_aware_training=True
    ).train()
    x = torch.from_numpy(rng.uniform(-1, 1, size=(10, 20, 32))).requires_grad_(True)
    y = layer(x)[:, -1]
    y_sat = layer_sat(x)[:, -1]
    y_ground = torch.nn.ReLU()(x)[:, -1]

    # forward pass is different
    assert allclose(y.detach().numpy(), y_ground.detach().numpy())
    assert not allclose(
        y_sat.detach().numpy(),
        y_ground.detach().numpy(),
        record_rmse=False,
        print_fail=0,
    )

    # gradients are the same
    dy = torch.autograd.grad(y, x, grad_outputs=[torch.ones_like(y)])[0]
    dy_ground = torch.autograd.grad(y_ground, x, grad_outputs=[torch.ones_like(y)])[0]
    dy_sat = torch.autograd.grad(y_sat, x, grad_outputs=[torch.ones_like(y)])[0]
    assert allclose(dy, dy_ground)
    assert allclose(dy_sat, dy_ground)


def test_spiking_swap_functional(allclose):
    class MyModel(torch.nn.Module):
        def __init__(self):
            super().__init__()

            self.dense0 = torch.nn.Linear(1, 10)
            self.dense1 = torch.nn.Linear(1, 10)

        def forward(self, inputs):
            x = inputs.view((-1, inputs.shape[-1]))

            x0 = self.dense0(x)
            x0 = x0.view((inputs.shape[0], inputs.shape[1], x0.shape[-1]))
            x0 = torch.nn.LeakyReLU(negative_slope=0.3)(x0)

            x1 = self.dense1(x)
            x1 = x1.view((inputs.shape[0], inputs.shape[1], x1.shape[-1]))
            x1 = modules.SpikingActivation(
                torch.nn.LeakyReLU(negative_slope=0.3),
                return_sequences=True,
                spiking_aware_training=False,
            )(x1)

            return x0, x1

    model = MyModel()
    loss_func = torch.nn.MSELoss(reduction="none")
    optimizer = torch.optim.SGD(model.parameters(), lr=0.1)

    for _ in range(200):
        model.zero_grad()

        outputs = model(torch.ones((32, 1, 1)))
        loss = sum(
            torch.mean(torch.sum(loss_func(o, t), dim=1))
            for o, t in zip(
                outputs,
                [torch.ones((32, 1, 10)) * torch.arange(1.0, 100.0, 10)] * 2,
            )
        )
        loss.backward()
        optimizer.step()

    y0, y1 = model(torch.ones((1, 1000, 1)))
    assert allclose(y0.detach().numpy(), np.arange(1, 100, 10), atol=1)
    assert allclose(
        np.sum(y1.detach().numpy() * 0.001, axis=1, keepdims=True),
        np.arange(1, 100, 10),
        atol=1,
    )


@pytest.mark.parametrize("dt", (0.001, 1))
def test_lowpass_tau(dt, allclose, rng):
    nengo = pytest.importorskip("nengo")

    # verify that the pytorch-spiking lowpass implementation matches the nengo lowpass
    # implementation
    layer = modules.Lowpass(tau=0.1, units=32, dt=dt).double()

    with torch.no_grad():
        x = torch.from_numpy(rng.randn(10, 100, 32))
        y = layer(x)

    y_nengo = nengo.Lowpass(0.1).filt(x, axis=1, dt=dt)

    assert allclose(y, y_nengo)


def test_lowpass_apply_during_training(allclose, rng):
    with torch.no_grad():
        x = torch.from_numpy(rng.randn(10, 100, 32))

        # apply_during_training=False:
        #   confirm `output == input` for training=True, but not training=False
        layer = modules.Lowpass(
            tau=0.1, units=32, apply_during_training=False, return_sequences=True
        )
        assert allclose(layer.train()(x), x)
        assert not allclose(layer.eval()(x), x, record_rmse=False, print_fail=0)

        # apply_during_training=True:
        #   confirm `output != input` for both values of `training`, and
        #   output is equal for both values of `training`
        layer = modules.Lowpass(
            tau=0.1, units=32, apply_during_training=True, return_sequences=True
        )
        assert not allclose(layer.train()(x), x, record_rmse=False, print_fail=0)
        assert not allclose(layer.eval()(x), x, record_rmse=False, print_fail=0)
        assert allclose(layer.train()(x), layer.eval()(x))


def test_lowpass_trainable(allclose):
    class MyModel(torch.nn.Module):
        def __init__(self):
            super().__init__()

            self.trained = modules.Lowpass(0.01, 1, apply_during_training=True)
            self.skip = modules.Lowpass(0.01, 1, apply_during_training=False)
            self.untrained = modules.Lowpass(0.01, 1, apply_during_training=True)
            for param in self.untrained.parameters():
                param.requires_grad = False

        def forward(self, inputs):
            return self.trained(inputs), self.skip(inputs), self.untrained(inputs)

    model = MyModel()

    loss_func = torch.nn.MSELoss(reduction="mean")
    optimizer = torch.optim.SGD(model.parameters(), lr=0.5)

    for _ in range(10):
        model.zero_grad()

        outputs = model(torch.zeros(1, 1, 1))
        loss = sum(
            loss_func(o, t)
            for o, t in zip(
                outputs,
                [torch.ones(1, 1)] * 3,
            )
        )
        loss.backward()
        optimizer.step()

    # trainable layer should learn to output 1
    ys = model(torch.zeros((1, 1, 1)))
    assert allclose(ys[0].detach(), 1)
    assert not allclose(ys[1].detach(), 1, record_rmse=False, print_fail=0)
    assert not allclose(ys[2].detach(), 1, record_rmse=False, print_fail=0)

    # for trainable layer, smoothing * initial_level should go to 1
    assert allclose(
        (torch.sigmoid(model.trained.smoothing_var) * model.trained.level_var).detach(),
        1,
    )

    # other layers should stay at initial value
    assert allclose(model.skip.level_var.detach(), 0)
    assert allclose(model.untrained.level_var.detach(), 0)
    assert allclose(model.skip.smoothing_var.detach(), model.skip.smoothing_init)
    assert allclose(
        model.untrained.smoothing_var.detach(), model.untrained.smoothing_init
    )


def test_lowpass_validation():
    with pytest.raises(ValueError, match="tau must be a positive number"):
        modules.Lowpass(tau=0, units=1)


def test_temporalavgpool(rng, allclose):
    x = rng.randn(32, 10, 2, 5)
    tx = torch.from_numpy(x)
    for dim in range(x.ndim):
        model = torch.nn.Sequential(modules.TemporalAvgPool(dim=dim))
        toutput = model(tx)
        assert allclose(toutput.numpy(), x.mean(axis=dim))


@pytest.mark.parametrize(
    "module",
    (
        modules.SpikingActivation(
            torch.nn.ReLU(), initial_state=torch.zeros((32, 50)), dt=1
        ),
        modules.Lowpass(tau=0.01, units=50, dt=0.001),
    ),
)
def test_return_sequences(module, rng, allclose):
    x = torch.tensor(rng.randn(32, 10, 50))

    with torch.no_grad():
        module_seq = copy.deepcopy(module)
        module_seq.return_sequences = True
        y_seq = module_seq(x)

        module_last = copy.deepcopy(module)
        module_last.return_sequences = False
        y_last = module_last(x)

    assert y_seq.shape == x.shape
    assert y_last.shape == (x.shape[0], x.shape[2])
    assert allclose(y_seq[:, -1], y_last)



# pylint: disable=missing-docstring

__copyright__ = "2020-2020, Applied Brain Research"
__license__ = "Free for non-commercial use; see LICENSE.rst"
from pytorch_spiking.version import version as __version__

from pytorch_spiking import modules, functional
from pytorch_spiking.modules import Lowpass, SpikingActivation, TemporalAvgPool
